#include "GLWidget.hh"

#include <QtCore/QCoreApplication>
#include <QtGui/QKeyEvent>
#include <QTimer>

#import <OpenGL/gl.h>
//#import <OpenGL/gl3.h>
/// @todo After we drop 10.6 support, switch back to gl3 and remove the below 4 lines.  See also r15430 for shader changes.
#include <OpenGL/glext.h>

#include <OpenGL/CGLCurrent.h>

#include <sys/time.h> // for gettimeofday
#include <dispatch/dispatch.h> // for USEC_PER_SEC


#define GLSL_STRING(version,source) "#version " #version "\n" #source

static const char * rootVertexShaderSource = GLSL_STRING(120,
	attribute vec2 position;
	varying vec2 texcoord;

	void main()
	{
		gl_Position = vec4(position, 0.0, 1.0);
		texcoord = position * vec2(0.5) + vec2(0.5);
	}
);

static const char * rootFragmentShaderSource = GLSL_STRING(120,
	uniform sampler2D texture;
	varying vec2 texcoord;

	void main()
	{
		gl_FragColor = texture2D(texture, texcoord);
	}
);

static const char * childVertexShaderSource = GLSL_STRING(120,
	attribute vec2 position;
	varying vec2 texcoord;

	void main()
	{
		gl_Position = vec4(position, 0.0, 1.0);
		texcoord = position * vec2(0.5) + vec2(0.5);
	}
);

static const char * childFragmentShaderSource = GLSL_STRING(120,
	uniform sampler2D texture;
	uniform float renderTime;
	varying vec2 texcoord;

	void main()
	{
		gl_FragColor = texture2D(texture, texcoord + vec2(sin((texcoord.x + renderTime)*3.14*2.) / 50., 0));
	}
);


static GLuint compileShader(GLenum type, const char * source)
{
	GLint length = strlen(source);
	GLuint shader = glCreateShader(type);
	glShaderSource(shader, 1, (const GLchar**)&source, &length);
	glCompileShader(shader);
	return shader;
}

static const GLfloat quadPositions[] = { -1, -1, 1, -1, -1, 1, 1, 1 };
static const GLushort quadElements[] = { 0, 1, 2, 3 };


GLWidget::GLWidget(QWidget* parent)
	: QGLWidget(parent)
{
}

void GLWidget::initializeGL()
{
	// Upload a quad, for rendering the texture later on
	glGenBuffers(1, &rootQuadPositionBuffer);
	glBindBuffer(GL_ARRAY_BUFFER, rootQuadPositionBuffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(quadPositions), quadPositions, GL_STATIC_DRAW);

	glGenBuffers(1, &rootQuadElementBuffer);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, rootQuadElementBuffer);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(quadElements), quadElements, GL_STATIC_DRAW);

	// Load an image from disk into a GL Texture
	QPixmap image(QCoreApplication::applicationDirPath() + "/../../../OttoOperatesTheRoller.jpg");
	rootInputTexture = bindTexture(image, GL_TEXTURE_2D);

	// Prepare a shader for displaying the GL Texture onscreen
	rootVertexShader = compileShader(GL_VERTEX_SHADER, rootVertexShaderSource);
	rootFragmentShader = compileShader(GL_FRAGMENT_SHADER, rootFragmentShaderSource);

	rootProgram = glCreateProgram();
	glAttachShader(rootProgram, rootVertexShader);
	glAttachShader(rootProgram, rootFragmentShader);
	glLinkProgram(rootProgram);

	rootPositionAttribute = glGetAttribLocation(rootProgram, "position");
	rootTextureUniform = glGetUniformLocation(rootProgram, "texture");



	// Create a shared context (so I can eventually move the framebuffer operation to another thread).
	rootContext = CGLGetCurrentContext();

	CGLPixelFormatObj pf = CGLGetPixelFormat(rootContext);
	CGLError error = CGLCreateContext(pf, rootContext, &childContext);
	if (error != kCGLNoError)
	{
		fprintf(stderr, "Error: failed to create childContext: %s\n", CGLErrorString(error));
		return;
	}



	// Create a framebuffer on the childContext, onto which I'll render the texture with fragment shader effects.
	CGLSetCurrentContext(childContext);
	{
		glGenTextures(1, &childTexture);
		glBindTexture(GL_TEXTURE_2D, childTexture);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, image.width(), image.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);

		glGenFramebuffers(1, &childFramebuffer);
		glBindFramebuffer(GL_FRAMEBUFFER, childFramebuffer);
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, childTexture, 0);

		glGenBuffers(1, &childQuadPositionBuffer);
		glBindBuffer(GL_ARRAY_BUFFER, childQuadPositionBuffer);
		glBufferData(GL_ARRAY_BUFFER, sizeof(quadPositions), quadPositions, GL_STATIC_DRAW);

		glGenBuffers(1, &childQuadElementBuffer);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, childQuadElementBuffer);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(quadElements), quadElements, GL_STATIC_DRAW);

		childVertexShader = compileShader(GL_VERTEX_SHADER, childVertexShaderSource);
		childFragmentShader = compileShader(GL_FRAGMENT_SHADER, childFragmentShaderSource);

		childProgram = glCreateProgram();
		glAttachShader(childProgram, childVertexShader);
		glAttachShader(childProgram, childFragmentShader);
		glLinkProgram(childProgram);

		childPositionAttribute = glGetAttribLocation(childProgram, "position");
		childTextureUniform = glGetUniformLocation(childProgram, "texture");
		childRenderTimeUniform = glGetUniformLocation(childProgram, "renderTime");
	}
	CGLSetCurrentContext(rootContext);



	QTimer *timer = new QTimer(this);
	connect(timer, SIGNAL(timeout()), this, SLOT(updateGL()));
	timer->start((int)(1000./60.));  // approximately 60 fps
}

void GLWidget::resizeGL(int w, int h)
{
	// Set the viewport to window dimensions
	glViewport( 0, 0, w, qMax( h, 1 ) );
}

void GLWidget::paintGL()
{
	// Execute the fragment shader every frame (on the childContext)
	//CGLSetCurrentContext(childContext);
	{
		glBindFramebuffer(GL_FRAMEBUFFER, childFramebuffer);
		{
			glUseProgram(childProgram);
			{
				struct timeval tv;
				gettimeofday(&tv,NULL);
				glUniform1f(childRenderTimeUniform, (float)tv.tv_usec/USEC_PER_SEC);

				glActiveTexture(GL_TEXTURE0);
				glBindTexture(GL_TEXTURE_2D, rootInputTexture);
				glUniform1i(childTextureUniform, 0);

				glBindBuffer(GL_ARRAY_BUFFER, childQuadPositionBuffer);
				glVertexAttribPointer(childPositionAttribute, 2, GL_FLOAT, GL_FALSE, sizeof(GLfloat)*2, (void*)0);
				glEnableVertexAttribArray(childPositionAttribute);
				{
					glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, childQuadElementBuffer);
					glDrawElements(GL_TRIANGLE_STRIP, 4, GL_UNSIGNED_SHORT, (void*)0);
				}
				glDisableVertexAttribArray(childPositionAttribute);
			}
			glUseProgram(0);
		}
		glBindFramebuffer(GL_FRAMEBUFFER, 0);
	}



	// Display the GL Texture onscreen (on the rootContext)
	CGLSetCurrentContext(rootContext);
	glUseProgram(rootProgram);
	{
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, childTexture);  // This texture was created and popupated on childContext.
//		glBindTexture(GL_TEXTURE_2D, inputTexture);
		glUniform1i(rootTextureUniform, 0);

		glBindBuffer(GL_ARRAY_BUFFER, rootQuadPositionBuffer);
		glVertexAttribPointer(rootPositionAttribute, 2, GL_FLOAT, GL_FALSE, sizeof(GLfloat)*2, (void*)0);
		glEnableVertexAttribArray(rootPositionAttribute);
		{
			glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, rootQuadElementBuffer);
			glDrawElements(GL_TRIANGLE_STRIP, 4, GL_UNSIGNED_SHORT, (void*)0);
		}
		glDisableVertexAttribArray(rootPositionAttribute);
	}
	glUseProgram(0);
}

void GLWidget::keyPressEvent(QKeyEvent* e)
{
	switch ( e->key() )
	{
		case Qt::Key_Escape:
			QCoreApplication::instance()->quit();
			break;

		default:
			QGLWidget::keyPressEvent( e );
	}
}
